@@ -105,11 +105,15 @@ You can cancel a function scheduled with any of these methods using
 Displaying FPS
 ==============
 
-The ClockDisplay class provides a simple FPS counter::
+The ClockDisplay class provides a simple FPS counter.  You should create
+an instance of ClockDisplay once during the application's start up::
 
     fps_display = clock.ClockDisplay()
+
+Call draw on the ClockDisplay object for each frame::
+
     fps_display.draw()
-   
+
 There are several options to change the font, color and text displayed
 within the __init__ method.
 
@@ -135,7 +139,7 @@ of the system clock.
 '''
 
 __docformat__ = 'restructuredtext'
-__version__ = '$Id: clock.py 1579 2008-01-15 14:47:19Z Alex.Holkner $'
+__version__ = '$Id: clock.py 2252 2008-09-08 05:53:44Z Alex.Holkner $'
 
 import time
 import sys
@@ -169,6 +173,13 @@ else:
         def sleep(self, microseconds):
             _c.usleep(int(microseconds))
 
+class _ScheduledItem(object):
+    __slots__ = ['func', 'args', 'kwargs']
+    def __init__(self, func, args, kwargs):
+        self.func = func
+        self.args = args
+        self.kwargs = kwargs
+
 class _ScheduledIntervalItem(object):
     __slots__ = ['func', 'interval', 'last_ts', 'next_ts', 
                  'args', 'kwargs']
@@ -180,6 +191,12 @@ class _ScheduledIntervalItem(object):
         self.args = args
         self.kwargs = kwargs
 
+def _dummy_schedule_func(*args, **kwargs):
+    '''Dummy function that does nothing, placed onto zombie scheduled items
+    to ensure they have no side effect if already queued inside tick() method.
+    '''
+    pass
+
 class Clock(_ClockBase):
     '''Class for calculating and limiting framerate, and for calling scheduled
     functions.
@@ -203,9 +220,6 @@ class Clock(_ClockBase):
     # List of schedule interval items kept in sort order.
     _schedule_interval_items = None
 
-    # Dict mapping function to schedule item for fast removal 
-    _schedule_functions = None
-
     # If True, a sleep(0) is inserted on every tick.   
     _force_sleep = False
 
@@ -216,7 +230,7 @@ class Clock(_ClockBase):
         :Parameters:
             `fps_limit` : float
                 If not None, the maximum allowable framerate.  Defaults
-                to None.
+                to None.  Deprecated in pyglet 1.1.
             `time_function` : function
                 Function to return the elapsed time of the application, 
                 in seconds.  Defaults to time.time, but can be replaced
@@ -235,22 +249,34 @@ class Clock(_ClockBase):
 
         self._schedule_items = []
         self._schedule_interval_items = []
-        self._schedule_functions = {}
 
-    def tick(self):
+    def tick(self, poll=False):
         '''Signify that one frame has passed.
 
         This will call any scheduled functions that have elapsed.
 
+        :Parameters:
+            `poll` : bool
+                If True, the function will call any scheduled functions
+                but will not sleep or busy-wait for any reason.  Recommended
+                for advanced applications managing their own sleep timers
+                only.
+                
+                Since pyglet 1.1.
+
         :rtype: float
         :return: The number of seconds since the last "tick", or 0 if this was
             the first frame.
         '''
-        if self.period_limit:
-            self._limit()
+        if poll:
+            if self.period_limit:
+                self.next_ts = self.next_ts + self.period_limit
+        else:
+            if self.period_limit:
+                self._limit()
 
-        if self._force_sleep:
-            self.sleep(0)
+            if self._force_sleep:
+                self.sleep(0)
 
         ts = self.time()
         if self.last_ts is None: 
@@ -265,8 +291,8 @@ class Clock(_ClockBase):
 
         # Call functions scheduled for every frame  
         # Dupe list just in case one of the items unchedules itself
-        for func, args, kwargs in list(self._schedule_items):
-            func(delta_t, *args, **kwargs)
+        for item in list(self._schedule_items):
+            item.func(delta_t, *item.args, **item.kwargs)
 
         # Call all scheduled interval functions and reschedule for future.
         need_resort = False
@@ -276,11 +302,23 @@ class Clock(_ClockBase):
                 break
             item.func(ts - item.last_ts, *item.args, **item.kwargs)
             if item.interval:
-                item.last_ts = ts
                 # Try to keep timing regular, even if overslept this time;
                 # but don't schedule in the past (which could lead to
                 # infinitely-worsing error).
-                item.next_ts = max(item.next_ts + item.interval, ts + 0.01)
+                item.next_ts = item.last_ts + item.interval
+                item.last_ts = ts
+                if item.next_ts <= ts:
+                    if ts - item.next_ts < 0.05:
+                        # Only missed by a little bit, keep the same schedule
+                        item.next_ts = ts + item.interval
+                    else:
+                        # Missed by heaps, do a soft reschedule to avoid 
+                        # lumping everything together.
+                        item.next_ts = self._get_soft_next_ts(ts, item.interval)
+                        # Fake last_ts to avoid repeatedly over-scheduling in
+                        # future.  Unfortunately means the next reported dt is
+                        # incorrect (looks like interval but actually isn't).
+                        item.last_ts = item.next_ts - item.interval
                 need_resort = True
 
         # Remove finished one-shots.
@@ -303,10 +341,10 @@ class Clock(_ClockBase):
         '''
         ts = self.time()
         # Sleep to just before the desired time
-        sleeptime = self.next_ts - self.time()
+        sleeptime = self.get_sleep_time(False)
         while sleeptime - self.SLEEP_UNDERSHOOT > self.MIN_SLEEP:
             self.sleep(1000000 * (sleeptime - self.SLEEP_UNDERSHOOT))
-            sleeptime = self.next_ts - self.time()
+            sleeptime = self.get_sleep_time(False)
 
         # Busy-loop CPU to get closest to the mark
         sleeptime = self.next_ts - self.time()
@@ -321,14 +359,65 @@ class Clock(_ClockBase):
             # Otherwise keep the clock steady
             self.next_ts = self.next_ts + self.period_limit
 
+    def get_sleep_time(self, sleep_idle):
+        '''Get the time until the next item is scheduled.
+
+        This method considers all scheduled items and the current
+        ``fps_limit``, if any.
+
+        Applications can choose to continue receiving updates at the
+        maximum framerate during idle time (when no functions are scheduled),
+        or they can sleep through their idle time and allow the CPU to
+        switch to other processes or run in low-power mode.
+
+        If `sleep_idle` is ``True`` the latter behaviour is selected, and
+        ``None`` will be returned if there are no scheduled items.
+
+        Otherwise, if `sleep_idle` is ``False``, a sleep time allowing
+        the maximum possible framerate (considering ``fps_limit``) will
+        be returned; or an earlier time if a scheduled function is ready.
+
+        :Parameters:
+            `sleep_idle` : bool
+                If True, the application intends to sleep through its idle
+                time; otherwise it will continue ticking at the maximum 
+                frame rate allowed.
+
+        :rtype: float
+        :return: Time until the next scheduled event in seconds, or ``None``
+            if there is no event scheduled.
+
+        :since: pyglet 1.1
+        '''
+        if self._schedule_items or not sleep_idle:
+            if not self.period_limit:
+                return 0.
+            else:
+                wake_time = self.next_ts
+                if self._schedule_interval_items:
+                    wake_time = min(wake_time,
+                                    self._schedule_interval_items[0].next_ts)
+                return max(wake_time - self.time(), 0.)
+
+        if self._schedule_interval_items:
+            return max(self._schedule_interval_items[0].next_ts - self.time(),
+                       0)
+            
+        return None
+
     def set_fps_limit(self, fps_limit):
         '''Set the framerate limit.
 
+        The framerate limit applies only when a function is scheduled
+        for every frame.  That is, the framerate limit can be exceeded by
+        scheduling a function for a very small period of time.
+
         :Parameters:
             `fps_limit` : float
                 Maximum frames per second allowed, or None to disable
                 limiting.
 
+        :deprecated: Use `pyglet.app.run` and `schedule_interval` instead.
         '''
         if not fps_limit:
             self.period_limit = None
@@ -376,26 +465,10 @@ class Clock(_ClockBase):
             `func` : function
                 The function to call each frame.
         '''
-        item = (func, args, kwargs)
+        item = _ScheduledItem(func, args, kwargs)
         self._schedule_items.append(item)
 
-        # _schedule_functions gives mapping of func to item.  if func
-        # is already scheduled, the mapping becomes a list of items.
-        if func in self._schedule_functions:
-            entry = self._schedule_functions[func]
-            if type(entry) == list:
-                self._schedule_functions[func].append(item)
-            else:
-                self._schedule_functions[func] = [entry, item]
-        else:
-            self._schedule_functions[func] = item
-
-    def _schedule_item(self, func, interval, repeat, *args, **kwargs):
-        last_ts = self.last_ts or self.next_ts
-        next_ts = last_ts + interval
-        if not repeat:
-            interval = 0
-
+    def _schedule_item(self, func, last_ts, next_ts, interval, *args, **kwargs):
         item = _ScheduledIntervalItem(
             func, interval, last_ts, next_ts, args, kwargs)
 
@@ -407,16 +480,6 @@ class Clock(_ClockBase):
         else:
             self._schedule_interval_items.append(item)
 
-        # add item to func mapping
-        if func in self._schedule_functions:
-            entry = self._schedule_functions[func]
-            if type(entry) == list:
-                self._schedule_functions[func].append(item)
-            else:
-                self._schedule_functions[func] = [entry, item]
-        else:
-            self._schedule_functions[func] = item
-
     def schedule_interval(self, func, interval, *args, **kwargs):
         '''Schedule a function to be called every `interval` seconds.
 
@@ -432,7 +495,108 @@ class Clock(_ClockBase):
                 The number of seconds to wait between each call.
 
         '''
-        self._schedule_item(func, interval, True, *args, **kwargs)
+        last_ts = self.last_ts or self.next_ts
+
+        # Schedule from now, unless now is sufficiently close to last_ts, in
+        # which case use last_ts.  This clusters together scheduled items that
+        # probably want to be scheduled together.  The old (pre 1.1.1)
+        # behaviour was to always use self.last_ts, and not look at ts.  The
+        # new behaviour is needed because clock ticks can now be quite
+        # irregular, and span several seconds.
+        ts = self.time()
+        if ts - last_ts > 0.2:
+            last_ts = ts
+
+        next_ts = last_ts + interval
+        self._schedule_item(func, last_ts, next_ts, interval, *args, **kwargs)
+
+    def schedule_interval_soft(self, func, interval, *args, **kwargs):
+        '''Schedule a function to be called every `interval` seconds,
+        beginning at a time that does not coincide with other scheduled
+        events.
+        
+        This method is similar to `schedule_interval`, except that the
+        clock will move the interval out of phase with other scheduled
+        functions so as to distribute CPU more load evenly over time.
+
+        This is useful for functions that need to be called regularly, 
+        but not relative to the initial start time.  `pyglet.media`
+        does this for scheduling audio buffer updates, which need to occur
+        regularly -- if all audio updates are scheduled at the same time 
+        (for example, mixing several tracks of a music score, or playing
+        multiple videos back simultaneously), the resulting load on the
+        CPU is excessive for those intervals but idle outside.  Using
+        the soft interval scheduling, the load is more evenly distributed.
+
+        Soft interval scheduling can also be used as an easy way to schedule
+        graphics animations out of phase; for example, multiple flags
+        waving in the wind.
+
+        :since: pyglet 1.1
+
+        :Parameters:
+            `func` : function
+                The function to call when the timer lapses.
+            `interval` : float
+                The number of seconds to wait between each call.
+
+        '''
+        last_ts = self.last_ts or self.next_ts
+
+        # See schedule_interval
+        ts = self.time()
+        if ts - last_ts > 0.2:
+            last_ts = ts
+
+        next_ts = self._get_soft_next_ts(last_ts, interval)
+        last_ts = next_ts - interval
+        self._schedule_item(func, last_ts, next_ts, interval, *args, **kwargs)
+
+    def _get_soft_next_ts(self, last_ts, interval):
+        def taken(ts, e):
+            '''Return True if the given time has already got an item
+            scheduled nearby.
+            '''
+            for item in self._schedule_interval_items:
+                if abs(item.next_ts - ts) <= e:
+                    return True
+                elif item.next_ts > ts + e:
+                    return False
+            return False
+
+        # Binary division over interval:
+        #
+        # 0                          interval
+        # |--------------------------|
+        #   5  3   6   2   7  4  8   1          Order of search
+        #
+        # i.e., first scheduled at interval,
+        #       then at            interval/2 
+        #       then at            interval/4 
+        #       then at            interval*3/4
+        #       then at            ...
+        #
+        # Schedule is hopefully then evenly distributed for any interval,
+        # and any number of scheduled functions.
+
+        next_ts = last_ts + interval
+        if not taken(next_ts, interval / 4):
+            return next_ts
+
+        dt = interval
+        divs = 1
+        while True:
+            next_ts = last_ts
+            for i in range(divs - 1):
+                next_ts += dt
+                if not taken(next_ts, dt / 4):
+                    return next_ts
+            dt /= 2
+            divs *= 2
+            
+            # Avoid infinite loop in pathological case
+            if divs > 16:
+                return next_ts
 
     def schedule_once(self, func, delay, *args, **kwargs):
         '''Schedule a function to be called once after `delay` seconds.
@@ -445,7 +609,15 @@ class Clock(_ClockBase):
             `delay` : float
                 The number of seconds to wait before the timer lapses.
         '''
-        self._schedule_item(func, delay, False, *args, **kwargs)
+        last_ts = self.last_ts or self.next_ts
+
+        # See schedule_interval
+        ts = self.time()
+        if ts - last_ts > 0.2:
+            last_ts = ts
+
+        next_ts = last_ts + delay
+        self._schedule_item(func, last_ts, next_ts, 0, *args, **kwargs)
 
     def unschedule(self, func):
         '''Remove a function from the schedule.  
@@ -458,23 +630,25 @@ class Clock(_ClockBase):
                 The function to remove from the schedule.
 
         '''
-        if func not in self._schedule_functions:
-            return
+        # First replace zombie items' func with a dummy func that does
+        # nothing, in case the list has already been cloned inside tick().
+        # (Fixes issue 326).
+        for item in self._schedule_items:
+            if item.func == func:
+                item.func = _dummy_schedule_func
 
-        items = self._schedule_functions[func]
-        if type(items) == list:
-            for item in items:
-                if item in self._schedule_items:
-                    self._schedule_items.remove(item)
-                elif item in self._schedule_interval_items:
-                    self._schedule_interval_items.remove(item)
-        else:
-            if items in self._schedule_items:
-                self._schedule_items.remove(items)
-            elif items in self._schedule_interval_items:
-                self._schedule_interval_items.remove(items)
-        del self._schedule_functions[func]
+        for item in self._schedule_interval_items:
+            if item.func == func:
+                item.func = _dummy_schedule_func
+    
+        # Now remove matching items from both schedule lists.
+        self._schedule_items = \
+            [item for item in self._schedule_items \
+                  if item.func is not _dummy_schedule_func]
 
+        self._schedule_interval_items = \
+            [item for item in self._schedule_interval_items \
+                  if item.func is not _dummy_schedule_func]
 
 # Default clock.
 _default = Clock()
@@ -500,17 +674,45 @@ def get_default():
     '''
     return _default
 
-def tick():
+def tick(poll=False):
     '''Signify that one frame has passed on the default clock.
 
     This will call any scheduled functions that have elapsed.
 
+    :Parameters:
+        `poll` : bool
+            If True, the function will call any scheduled functions
+            but will not sleep or busy-wait for any reason.  Recommended
+            for advanced applications managing their own sleep timers
+            only.
+            
+            Since pyglet 1.1.
+
     :rtype: float
     :return: The number of seconds since the last "tick", or 0 if this was the
         first frame.
     '''
 
-    return _default.tick()
+    return _default.tick(poll)
+
+def get_sleep_time(sleep_idle):
+    '''Get the time until the next item is scheduled on the default clock.
+
+    See `Clock.get_sleep_time` for details.
+
+    :Parameters:
+        `sleep_idle` : bool
+            If True, the application intends to sleep through its idle
+            time; otherwise it will continue ticking at the maximum 
+            frame rate allowed.
+
+    :rtype: float
+    :return: Time until the next scheduled event in seconds, or ``None``
+        if there is no event scheduled.
+
+    :since: pyglet 1.1
+    '''
+    return _default.get_sleep_time(sleep_idle)
 
 def get_fps():
     '''Return the current measured FPS of the default clock.
@@ -527,6 +729,7 @@ def set_fps_limit(fps_limit):
             Maximum frames per second allowed, or None to disable
             limiting.
 
+    :deprecated: Use `pyglet.app.run` and `schedule_interval` instead.
     '''
     _default.set_fps_limit(fps_limit)
 
@@ -567,6 +770,27 @@ def schedule_interval(func, interval, *args, **kwargs):
     '''
     _default.schedule_interval(func, interval, *args, **kwargs)
 
+def schedule_interval_soft(func, interval, *args, **kwargs):
+    '''Schedule 'func' to be called every 'interval' seconds on the default
+    clock, beginning at a time that does not coincide with other scheduled
+    events. 
+    
+    The arguments passed to 'func' are 'dt' (time since last function call),
+    followed by any ``*args`` and ``**kwargs`` given here.
+
+    :see: `Clock.schedule_interval_soft`
+
+    :since: pyglet 1.1
+    
+    :Parameters:
+        `func` : function
+            The function to call when the timer lapses.
+        `interval` : float
+            The number of seconds to wait between each call.
+
+    '''
+    _default.schedule_interval_soft(func, interval, *args, **kwargs)
+
 def schedule_once(func, delay, *args, **kwargs):
     '''Schedule 'func' to be called once after 'delay' seconds (can be
     a float) on the default clock.  The arguments passed to 'func' are
@@ -649,6 +873,19 @@ class ClockDisplay(object):
 
         self.format = format
 
+    def unschedule(self):
+        '''Remove the display from its clock's schedule.
+
+        `ClockDisplay` uses `Clock.schedule_interval` to periodically update
+        its display label.  Even if the ClockDisplay is not being used any
+        more, its update method will still be scheduled, which can be a
+        resource drain.  Call this method to unschedule the update method
+        and allow the ClockDisplay to be garbage collected.
+
+        :since: pyglet 1.1
+        '''
+        self.clock.unschedule(self.update_text)
+
     def update_text(self, dt=0):
         '''Scheduled method to update the label text.''' 
         fps = self.clock.get_fps()
