@@ -31,7 +31,7 @@
 # ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 # POSSIBILITY OF SUCH DAMAGE.
 # ----------------------------------------------------------------------------
-# $Id: __init__.py 1579 2008-01-15 14:47:19Z Alex.Holkner $
+# $Id: __init__.py 2154 2008-08-05 22:53:37Z Alex.Holkner $
 
 '''Audio and video playback.
 
@@ -51,12 +51,7 @@ associated with a single player by "queuing" it::
     source = load('background_music.mp3')
     player.queue(source)
 
-Use the `Player` to control playback.  Within your main run loop, you must
-periodically call `dispatch_events` to ensure the audio buffers are refilled::
-
-    player.play()
-    while player.source:    # While the source hasn't finished
-        player.dispatch_events()
+Use the `Player` to control playback.  
 
 If the source contains video, the `Source.video_format` attribute will be
 non-None, and the `Player.texture` attribute will contain the current video
@@ -75,19 +70,23 @@ of players, and so played many times simultaneously.
 '''
 
 __docformat__ = 'restructuredtext'
-__version__ = '$Id: __init__.py 1579 2008-01-15 14:47:19Z Alex.Holkner $'
+__version__ = '$Id: __init__.py 2154 2008-08-05 22:53:37Z Alex.Holkner $'
 
 import ctypes
 import sys
 import time
 import StringIO
 
+import pyglet
+from pyglet import clock
 from pyglet import event
 
+_debug_media = pyglet.options['debug_media']
+
 class MediaException(Exception):
     pass
 
-class MediaFormatException(Exception):
+class MediaFormatException(MediaException):
     pass
 
 class CannotSeekException(MediaException):
@@ -105,7 +104,7 @@ class AudioFormat(object):
             The number of channels: 1 for mono or 2 for stereo (pyglet does
             not yet support surround-sound sources).
         `sample_size` : int
-            Bits per sample; typically 8 or 16.
+            Bits per sample; only 8 or 16 are supported.
         `sample_rate` : int
             Samples per second (in Hertz).
 
@@ -218,8 +217,9 @@ class AudioPlayer(object):
     '''Abstract low-level interface for playing audio.
 
     AudioPlayer has no knowledge of sources or eos behaviour.  Once
-    created, its audio format cannot be modified.  Behaviour is undefined if
-    there is a data underrun.
+    created, its audio format cannot be modified.  The player will attempt
+    to recover automatically from a buffer underrun (but this is not
+    guaranteed).
     
     Applications should not use this class directly, but instead use `Player`.
 
@@ -228,6 +228,8 @@ class AudioPlayer(object):
             The player's audio format (read-only).
 
     '''
+
+    UPDATE_PERIOD = 0.15
     
     def __init__(self, audio_format):
         '''Create a new audio player for the given audio format.
@@ -395,6 +397,64 @@ class Source(object):
         player.play()
         return player
 
+    def get_animation(self):
+        '''Import all video frames into memory as an `Animation`.
+
+        An empty animation will be returned if the source has no video.
+        Otherwise, the animation will contain all unplayed video frames (the
+        entire source, if it has not been queued on a player).  After creating
+        the animation, the source will be at EOS.
+
+        This method is unsuitable for videos running longer than a
+        few seconds.
+
+        :since: pyglet 1.1
+
+        :rtype: `pyglet.image.Animation`
+        '''
+        from pyglet.image import Animation, AnimationFrame
+        if not self.video_format:
+            return Animation([])
+        else:
+            # Create a dummy player for the source to push its textures onto.
+            frames = []
+            last_ts = 0
+            next_ts = self.get_next_video_timestamp()
+            while next_ts is not None:
+                image = self.get_next_video_frame()
+                assert image is not None
+                delay = next_ts - last_ts
+                frames.append(AnimationFrame(image, delay))
+                last_ts = next_ts
+                next_ts = self.get_next_video_timestamp()
+            return Animation(frames)
+
+    def get_next_video_timestamp(self):
+        '''Get the timestamp of the next video frame.
+
+        :since: pyglet 1.1
+
+        :rtype: float
+        :return: The next timestamp, or ``None`` if there are no more video
+            frames.
+        '''
+        pass
+
+    def get_next_video_frame(self):
+        '''Get the next video frame.
+
+        Video frames may share memory: the previous frame may be invalidated
+        or corrupted when this method is called unless the application has
+        made a copy of it.
+
+        :since: pyglet 1.1
+
+        :rtype: `pyglet.image.AbstractImage`
+        :return: The next video frame image, or ``None`` if there are no more
+            video frames.
+        '''
+        pass
+
     # Internal methods that Players call on the source:
 
     def _play(self):
@@ -435,7 +495,7 @@ class Source(object):
     def _init_texture(self, player):
         '''Create the player's texture.'''
         pass
-
+    
     def _update_texture(self, player, timestamp):
         '''Update the texture on player.'''
         pass
@@ -601,7 +661,7 @@ class Player(event.EventDispatcher):
     _position = (0, 0, 0)
     _pitch = 1.0
 
-    _cone_orientation = (0, 0, 0)
+    _cone_orientation = (0, 0, 1)
     _cone_inner_angle = 360.
     _cone_outer_angle = 360.
     _cone_outer_gain = 1.
@@ -630,6 +690,15 @@ class Player(event.EventDispatcher):
                 self._audio = None
 
         self._audio = audio_player_class(source.audio_format)
+        self._audio.set_volume(self._volume)
+        self._audio.set_min_distance(self._min_distance)
+        self._audio.set_max_distance(self._max_distance)
+        self._audio.set_position(self._position)
+        self._audio.set_pitch(self._pitch)
+        self._audio.set_cone_orientation(self._cone_orientation)
+        self._audio.set_cone_inner_angle(self._cone_inner_angle)
+        self._audio.set_cone_outer_angle(self._cone_outer_angle)
+        self._audio.set_cone_outer_gain(self._cone_outer_gain)
 
     def _fill_audio(self):
         '''Ensure _audio is full.'''
@@ -701,6 +770,16 @@ class Player(event.EventDispatcher):
         if not source:
             yield 'end', None
 
+    def _update_schedule(self):
+        clock.unschedule(self.dispatch_events)
+        if self._playing and self._sources:
+            interval = 1000.
+            if self._sources[0].video_format:
+                interval = min(interval, 1/24.)
+            if self._audio:
+                interval = min(interval, self._audio.UPDATE_PERIOD)
+            clock.schedule_interval_soft(self.dispatch_events, interval)
+
     def queue(self, source):
         '''Queue the source on this player.
 
@@ -731,6 +810,9 @@ class Player(event.EventDispatcher):
         else:
             self._last_system_time = time.time()
 
+        self.dispatch_events()
+        self._update_schedule()
+
     def pause(self):
         '''Pause playback of the current source.
 
@@ -741,6 +823,7 @@ class Player(event.EventDispatcher):
         
         if self._audio:
             self._audio.stop()
+        self._update_schedule()
 
     def seek(self, timestamp):
         '''Seek for playback to the indicated timestamp in seconds on the
@@ -752,7 +835,7 @@ class Player(event.EventDispatcher):
                 Timestamp to seek to.
         '''
         if not self._sources:
-            pass
+            return
 
         if not self._playing:
             self._pause_seek = True
@@ -760,6 +843,7 @@ class Player(event.EventDispatcher):
         self._audio_finished = False
         source = self._sources[0]
         self._source_read_index = 0
+        self._next_audio_data = None
         source._seek(timestamp)
         self._timestamp = timestamp
 
@@ -768,6 +852,8 @@ class Player(event.EventDispatcher):
             self._audio.clear()
         else:
             self._last_system_time = time.time()
+
+        self.dispatch_events()
         
     def next(self):
         '''Move immediately to the next queued source.
@@ -788,9 +874,10 @@ class Player(event.EventDispatcher):
 
     def _next_source(self):
         if not self._sources:
+            self._update_schedule()
             return
 
-        self._source_read_index -= 1
+        self._source_read_index = max(0, self._source_read_index - 1)
         source = self._sources.pop(0)
         source._release_texture(self)
         source._stop()
@@ -810,6 +897,7 @@ class Player(event.EventDispatcher):
 
         if self._playing:
             self.play()
+            self._update_schedule()
 
     def _on_eos(self):
         '''Internal method when EOS is encountered.  Returns False if
@@ -826,9 +914,18 @@ class Player(event.EventDispatcher):
         self.dispatch_event('on_eos')
         return True
 
-    def dispatch_events(self):
+    def dispatch_events(self, dt=None):
         '''Dispatch any pending events and perform regular heartbeat functions
         to maintain playback.
+
+        :Parameters:
+            `dt` : None
+                Ignored (for compatibility with `pyglet.clock.schedule`)
+
+        :deprecated: Since pyglet 1.1, Player objects schedule themselves on
+            the default clock automatically.  Applications should not call
+            this method.
+
         '''
         if not self._sources:
             return
@@ -837,10 +934,17 @@ class Player(event.EventDispatcher):
             self._fill_audio()
 
         if self._audio:
-            self._audio.pump()
+            underrun = self._audio.pump()
             while self._audio.clear_eos():
                 if not self._on_eos():
                     return
+            if underrun:
+                self._audio.UPDATE_PERIOD *= 0.75
+                self._audio.__class__.UPDATE_PERIOD *= 0.75
+                self._update_schedule()
+                if _debug_media:
+                    print '%r underrun: reducing update period to %.2f' % \
+                        (self._audio, self._audio.UPDATE_PERIOD)
         else:
             if self._playing:
                 t = time.time()
@@ -1070,6 +1174,19 @@ class Player(event.EventDispatcher):
         :type: float
         ''')
 
+    def get_texture(self):
+        '''Get the texture for the current video frame.
+
+        You should call this method every time you display a frame
+        of video, as multiple textures might be used.  The return value will
+        be `None` if there is no video in the current source.
+
+        :since: pyglet 1.1
+
+        :rtype: `pyglet.image.Texture`
+        '''
+        return self._texture
+
     texture = property(lambda self: self._texture,
                        doc='''The video texture.
 
@@ -1077,6 +1194,8 @@ class Player(event.EventDispatcher):
         of video, as multiple textures might be used.  This property will
         be `None` if there is no video in the current source.
 
+        :deprecated: Use `get_texture`.
+
         :type: `pyglet.image.Texture`
         ''')
 
@@ -1098,8 +1217,6 @@ class ManagedSoundPlayer(Player):
 
     This player will continue playing the sound until the sound is
     finished, even if the application discards the player early.
-    There is no need to call `Player.dispatch_events` on this player,
-    though you must call `pyglet.media.dispatch_events`.
 
     Only one source can be queued on the player; the player will be 
     discarded when the source finishes.
@@ -1125,6 +1242,7 @@ class ManagedSoundPlayer(Player):
 
     def stop(self):
         self._timestamp = 0.
+        clock.unschedule(self.dispatch_events)
         managed_players.remove(self)
 
 class Listener(object):
@@ -1202,6 +1320,14 @@ if getattr(sys, 'is_epydoc', False):
     #:
     #: :type: `Listener`
     listener = Listener()
+
+    #: Indication of the presence of AVbin.  When `have_avbin` is ``True``
+    #: pyglet will be able to play back compressed media streams such as
+    #: MP3, OGG and various video formats.  If ``False`` only uncompressed
+    #: Wave files can be loaded.
+    #:
+    #: :type: bool
+    have_avbin = False
 else:
     # Find best available sound driver according to user preference
     import pyglet
@@ -1211,25 +1337,35 @@ else:
             driver_name = 'pyglet.media.drivers.' + driver_name
             __import__(driver_name)
             driver = sys.modules[driver_name]
+            driver.driver_init()
             break
-        except (ImportError, AttributeError):
+        except (ImportError, AttributeError, MediaException):
             pass
 
     if not driver:
         raise ImportError('No suitable audio driver could be loaded.')
 
-    driver.driver_init()
     audio_player_class = driver.driver_audio_player_class
     listener = driver.driver_listener
 
     # Find best available source loader
+    have_avbin = False
     try:
         from pyglet.media import avbin
         _source_class = avbin.AVbinSource
+        have_avbin = True
     except ImportError:
         from pyglet.media import riff
         _source_class = riff.WaveSource
 
+# Pretend to import some common audio drivers so that py2exe/py2app
+# are fooled into packagin them.
+if False:
+    import pyglet.media.drivers.silent
+    import pyglet.media.drivers.openal
+    import pyglet.media.drivers.directsound
+    import pyglet.media.drivers.alsa
+
 def load(filename, file=None, streaming=True):
     '''Load a source from a file.
 
@@ -1258,6 +1394,11 @@ def dispatch_events():
 
     You must call this function regularly (typically once per run loop
     iteration) in order to keep audio buffers of managed players full.
+
+    :deprecated: Since pyglet 1.1, Player objects schedule themselves on
+        the default clock automatically.  Applications should not call this
+        method.
+
     '''
     for player in managed_players:
         player.dispatch_events()
