@@ -125,8 +125,8 @@ _motion_map = {
     (key.END, False):       key.MOTION_END_OF_LINE,
     (key.PAGEUP, False):    key.MOTION_PREVIOUS_PAGE,
     (key.PAGEDOWN, False):  key.MOTION_NEXT_PAGE,
-    (key.PAGEUP, True):     key.MOTION_BEGINNING_OF_FILE,
-    (key.PAGEDOWN, True):   key.MOTION_END_OF_FILE,
+    (key.HOME, True):       key.MOTION_BEGINNING_OF_FILE,
+    (key.END, True):        key.MOTION_END_OF_FILE,
     (key.BACKSPACE, False): key.MOTION_BACKSPACE,
     (key.DELETE, False):    key.MOTION_DELETE,
 }
@@ -360,7 +360,8 @@ class Win32Context(gl.Context):
         self._context = wgl.wglCreateContext(window._dc)
         if self._share:
             assert self._share._context is not None
-            wgl.wglShareLists(self._share._context, self._context)
+            if not wgl.wglShareLists(self._share._context, self._context):
+                raise gl.ContextException('Unable to share contexts')
 
     def destroy(self):
         super(Win32Context, self).destroy()
@@ -373,7 +374,8 @@ class Win32ContextARB(Win32Context):
         self._context = wgl.wglCreateContext(window._dc)
         if self._share:
             assert self._share._context is not None
-            wgl.wglShareLists(self._share._context, self._context)
+            if not wgl.wglShareLists(self._share._context, self._context):
+                raise gl.ContextException('Unable to share contexts')
 
 class Win32MouseCursor(MouseCursor):
     drawable = False
@@ -527,6 +529,8 @@ class Win32Window(BaseWindow):
 
     def close(self):
         super(Win32Window, self).close()
+        if not self._hwnd:
+            return
         _user32.DestroyWindow(self._hwnd)
         _user32.UnregisterClassW(self._window_class.lpszClassName, 0)
         self.set_mouse_platform_visible(True)
@@ -746,8 +750,8 @@ class Win32Window(BaseWindow):
 
         def get_icon(image):
             # Alpha-blended icon: see http://support.microsoft.com/kb/318876
-            image.format = 'BGRA'
-            image.pitch = len(image.format) * image.width
+            format = 'BGRA'
+            pitch = len(format) * image.width
 
             header = BITMAPV5HEADER()
             header.bV5Size = sizeof(header)
@@ -767,7 +771,8 @@ class Win32Window(BaseWindow):
                 byref(dataptr), None, 0)
             _user32.ReleaseDC(None, hdc)
 
-            memmove(dataptr, image.data, len(image.data))
+            data = image.get_data(format, pitch)
+            memmove(dataptr, data, len(data))
 
             mask = _gdi32.CreateBitmap(image.width, image.height, 1, 1, None)
 
@@ -818,6 +823,15 @@ class Win32Window(BaseWindow):
 
     def dispatch_events(self):
         self._allow_dispatch_event = True
+        self.dispatch_pending_events()
+
+        msg = MSG()
+        while _user32.PeekMessageW(byref(msg), 0, 0, 0, PM_REMOVE):
+            _user32.TranslateMessage(byref(msg))
+            _user32.DispatchMessageW(byref(msg))
+        self._allow_dispatch_event = False
+
+    def dispatch_pending_events(self):
         while self._event_queue:
             event = self._event_queue.pop(0)
             if type(event[0]) is str:
@@ -827,17 +841,11 @@ class Win32Window(BaseWindow):
                 # win32 event
                 event[0](*event[1:])
 
-        msg = MSG()
-        while _user32.PeekMessageW(byref(msg), 0, 0, 0, PM_REMOVE):
-            _user32.TranslateMessage(byref(msg))
-            _user32.DispatchMessageW(byref(msg))
-        self._allow_dispatch_event = False
-
     def _wnd_proc(self, hwnd, msg, wParam, lParam):
         event_handler = self._event_handlers.get(msg, None)
         result = 0
         if event_handler:
-            if self._allow_dispatch_event:
+            if self._allow_dispatch_event or not self._enable_event_queue:
                 result = event_handler(msg, wParam, lParam)
             else:
                 self._event_queue.append((event_handler, msg, wParam, lParam))
@@ -964,6 +972,11 @@ class Win32Window(BaseWindow):
             track.hwndTrack = self._hwnd
             _user32.TrackMouseEvent(byref(track))
 
+        # Don't generate motion/drag events when mouse hasn't moved. (Issue
+        # 305)
+        if self._mouse_x == x and self._mouse_y == y:
+            return 0
+
         self._mouse_x = x
         self._mouse_y = y
         
@@ -1050,14 +1063,40 @@ class Win32Window(BaseWindow):
         self.dispatch_event('on_close')
         return 0
 
+    def _immediate_redraw(self):
+        '''If using EventLoop, redraw and flip the window immediately.
+
+        Assumes window has GL context.
+        '''
+        from pyglet import app
+        if app.event_loop is not None:
+            self.dispatch_event('on_draw')
+            self.flip()
+
     @Win32EventHandler(WM_PAINT)
     def _event_paint(self, msg, wParam, lParam):
         self.dispatch_event('on_expose')
+
+        self.switch_to()
+        self._immediate_redraw()
+        
         # Validating the window using ValidateRect or ValidateRgn
         # doesn't clear the paint message when more than one window
         # is open [why?]; defer to DefWindowProc instead.
         return None
 
+    @Win32EventHandler(WM_SIZING)
+    def _event_sizing(self, msg, wParam, lParam):
+        rect = cast(lParam, POINTER(RECT)).contents
+        width, height = self.get_size()
+        self.switch_to()
+        self.dispatch_event('on_resize', width, height)
+
+        from pyglet import app
+        if app.event_loop is not None:
+            app.event_loop._idle_chance()
+        return 1
+
     @Win32EventHandler(WM_SIZE)
     def _event_size(self, msg, wParam, lParam):
         if not self._dc:
@@ -1076,7 +1115,19 @@ class Win32Window(BaseWindow):
             self.dispatch_event('on_show')
         w, h = self._get_location(lParam)
         self._reset_exclusive_mouse_screen()
+        self.switch_to()
         self.dispatch_event('on_resize', w, h)
+        self._immediate_redraw()
+        return 0
+
+    @Win32EventHandler(WM_SYSCOMMAND)
+    def _event_syscommand(self, msg, wParam, lParam):
+        if wParam & 0xfff0 in (SC_MOVE, SC_SIZE):
+            # Should be in WM_ENTERSIZEMOVE, but we never get that message.
+            from pyglet import app
+            if app.event_loop is not None:
+                app.event_loop._allow_polling = False
+                app.event_loop._idle_chance()
         return 0
 
     @Win32EventHandler(WM_MOVE)
@@ -1086,6 +1137,13 @@ class Win32Window(BaseWindow):
         self.dispatch_event('on_move', x, y)
         return 0
 
+    @Win32EventHandler(WM_EXITSIZEMOVE)
+    def _event_entersizemove(self, msg, wParam, lParam):
+        from pyglet import app
+        if app.event_loop is not None:
+            app.event_loop._allow_polling = True
+        return 0
+
     '''
     # Alternative to using WM_SETFOCUS and WM_KILLFOCUS.  Which
     # is better?
@@ -1130,4 +1188,4 @@ class Win32Window(BaseWindow):
     @Win32EventHandler(WM_ERASEBKGND)
     def _event_erasebkgnd(self, msg, wParam, lParam):
         # Prevent flicker during resize.
-        return 0
+        return 1
